import argparse
import typing

from typing_extensions import override

from approvaltests import (
    approvals,
    find_formatter_for_specified_class,
    register_formatter,
    verify,
)
from approvaltests.core.format_wrapper import FormatWrapper
from approvaltests.verifiable_objects.formatter_of_argparse_namespace import (
    FormatterOfArgparseNamespace,
)


def test_argparse_namespace() -> None:
    approvals.settings().allow_multiple_verify_calls_for_this_method()
    args = argparse.ArgumentParser()
    args.add_argument("foo")
    args.add_argument("--foo2")
    result = args.parse_args(["bar", "--foo2=bar2"])
    verify(FormatterOfArgparseNamespace(result))
    verify(result)


def test_register_formatter() -> None:
    class ExampleFormatterWrapper(FormatWrapper):
        @override
        def is_match(self, data: typing.Any) -> bool:
            return True

        @override
        def wrap(self, data: typing.Any) -> typing.Any:
            return 42

    with register_formatter(ExampleFormatterWrapper()):
        assert 42 == find_formatter_for_specified_class("Some Result")
        verify("Some Result")
    assert "Some Result" == find_formatter_for_specified_class("Some Result")



"""Approvaltests.scrubbers module."""

from .scrubbers import (
    combine_scrubbers,
    create_regex_scrubber,
    scrub_all_dates,
    scrub_all_guids,
    templates_regex_scrubber_with_lambda,
    templates_regex_scrubber_with_replacement,
)



import argparse

from approvaltests import verify_argument_parser


def test_argument_parser() -> None:
    parser = argparse.ArgumentParser(
        prog="my_program.py",
        description="My Description",
    )
    parser.add_argument("1st_argument", help="1st argument help text")
    parser.add_argument("--optional_argument", help="An Optional Argument help text")
    parser.add_argument("long_argument", help=f"{'Very' * 100} Long message")
    verify_argument_parser(parser)



import textwrap
from typing import List, Tuple

from approval_utilities.utilities.multiline_string_utils import remove_indentation_from
from approvaltests.scrubbers import create_regex_scrubber
from approvaltests.scrubbers.scrubbers import Scrubber

_custom_scrubbers = []


class DateScrubber:
    @staticmethod
    def add_scrubber(example: str, regex: str, *, display_message: bool = True) -> None:
        import re

        try:
            if not re.search(regex, example):
                raise Exception(f"Regex '{regex}' does not match example '{example}'")
        except re.error as e:
            raise Exception(f"Invalid regex pattern '{regex}': {e}")
        _custom_scrubbers.append((regex, [example]))

        if display_message:
            message = remove_indentation_from(
                f"""
                You are using a custom date scrubber. If you think the format you want to scrub would be useful for others, please add it to https://github.com/approvals/ApprovalTests.Python/issues/124.

                To suppress this message, use
                    DateScrubber.add_scrubber("{example}", "{regex}", display_message=False)
                """
            )
            print(message)

    @staticmethod
    def _clear_custom_scrubbers() -> None:
        """Clear all custom scrubbers. Used for testing purposes."""
        _custom_scrubbers.clear()

    @staticmethod
    def get_supported_formats() -> List[Tuple[str, List[str]]]:
        return [
            (
                "[a-zA-Z]{3} [a-zA-Z]{3} \\d{2} \\d{2}:\\d{2}:\\d{2}",
                ["Tue May 13 16:30:00"],
            ),
            (
                "[a-zA-Z]{3} [a-zA-Z]{3} \\d{2} \\d{2}:\\d{2}:\\d{2} [a-zA-Z]{3,4} \\d{4}",
                ["Wed Nov 17 22:28:33 EET 2021"],
            ),
            (
                "[a-zA-Z]{3} [a-zA-Z]{3} \\d{2} \\d{4} \\d{2}:\\d{2}:\\d{2}.\\d{3}",
                ["Tue May 13 2014 23:30:00.789"],
            ),
            (
                "[a-zA-Z]{3} [a-zA-Z]{3} \\d{2} \\d{2}:\\d{2}:\\d{2} -\\d{4} \\d{4}",
                ["Tue May 13 16:30:00 -0800 2014"],
            ),
            (
                "\\d{2} [a-zA-Z]{3} \\d{4} \\d{2}:\\d{2}:\\d{2},\\d{3}",
                ["13 May 2014 23:50:49,999"],
            ),
            (
                "[a-zA-Z]{3} \\d{2}, \\d{4} \\d{2}:\\d{2}:\\d{2} [a-zA-Z]{2} [a-zA-Z]{3}",
                ["May 13, 2014 11:30:00 PM PST"],
            ),
            ("\\d{2}:\\d{2}:\\d{2}", ["23:30:00"]),
            (
                "\\d{4}/\\d{2}/\\d{2} \\d{2}:\\d{2}:\\d{2}.\\d{2}\\d",
                ["2014/05/13 16:30:59.786"],
            ),
            (
                "\\d{4}-\\d{1,2}-\\d{1,2}T\\d{1,2}:\\d{2}Z",
                [
                    "2020-9-10T08:07Z",
                    "2020-09-9T08:07Z",
                    "2020-09-10T8:07Z",
                    "2020-09-10T08:07Z",
                ],
            ),
            (
                "\\d{4}-\\d{1,2}-\\d{1,2}T\\d{1,2}:\\d{2}:\\d{2}Z",
                ["2020-09-10T08:07:89Z"],
            ),
            (
                "\\d{4}-\\d{1,2}-\\d{1,2}T\\d{1,2}:\\d{2}\\:\\d{2}\\.\\d{3}Z",
                ["2020-09-10T01:23:45.678Z"],
            ),
            (
                r"\d{4}-\d{1,2}-\d{1,2}(?:T| )\d{1,2}:\d{2}:\d{2}\.\d{6}",
                ["2023-07-16 17:39:03.293919", "2023-12-06T11:59:47.090226"],
            ),
            ("\\d{8}T\\d{6}Z", ["20210505T091112Z"]),
            (
                r"(Mon|Tue|Wed|Thu|Fri|Sat|Sun)\s(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s([0-3]?\d)\s([0-1]\d:[0-5]\d:[0-5]\d)\s(\d{4})",
                ["Tue May 13 16:30:00 2014", "Wed Dec 11 14:59:44 2024"],
            ),
            (
                r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\+\d{2}:\d{2}",
                ["2021-09-10T08:07:00+03:00", "2021-01-01T00:00:00+00:00"],
            ),
            (
                r"[12]\d{3}[01]\d[0-3]\d_[0-2]\d[0-5]\d[0-5]\d",
                ["20250527_125703"],
            ),
            (
                r"\d{4}-\d{2}-\d{2}",
                ["2020-02-02"],
            ),
        ]

    def __init__(self, date_regex: str):
        self.date_regex = date_regex

    def scrub(self, date_str: str) -> str:
        return create_regex_scrubber(self.date_regex, lambda t: f"<date{t}>")(date_str)

    @staticmethod
    def get_scrubber_for(example: str) -> Scrubber:
        all_formats = DateScrubber.get_supported_formats() + _custom_scrubbers
        supported = ""
        for date_regex, examples in all_formats:
            supported += f"    {examples[0]} | {date_regex} \n"
            scrubber = DateScrubber(date_regex)
            if scrubber.scrub(example) == "<date0>":
                return scrubber.scrub

        raise Exception(
            textwrap.dedent(
                f"""\
                No date scrubber found for '{example}'. You can add a scrubber with:

                   DateScrubber.add_scrubber("{example}", "<your_regex_here>")

                For more help, see https://github.com/approvals/ApprovalTests.Python/blob/main/docs/how_to/scrub_dates.md.
                """
            )
        )



from typing import List, Optional


class GenericDiffReporterConfig:
    def __init__(self, name: str, path: str, extra_args: Optional[List[str]] = None):
        self.name = name
        self.path = path
        self.extra_args = extra_args or []

    def serialize(self) -> List[str]:
        result = [self.name, self.path]
        if self.extra_args:
            result.append(self.extra_args)
        return result


def create_config(config: List[str]) -> GenericDiffReporterConfig:
    return GenericDiffReporterConfig(*config)



from .generic_diff_reporter import GenericDiffReporter, create_config


class ReportWithVSCodeMacOS(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=create_config(
                [
                    "ReportWithVSCode",
                    "/Applications/Visual Studio Code.app/contents/Resources/app/bin/code",
                    ["-d"],
                ]
            )
        )


class ReportWithVSCode(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(config=create_config(["ReportWithVSCode", "code", ["-d"]]))



from approvaltests.internals.logs.log_commons import APPROVAL_TESTS_TEMP_DIRECTORY


def test_temp_directory_is_setup_automatically() -> None:
    # setup happens automatically
    gitignore_path = APPROVAL_TESTS_TEMP_DIRECTORY / ".gitignore"
    contents = gitignore_path.read_text()
    assert contents == "*"



import os
import sys
from difflib import unified_diff

from typing_extensions import override

from approval_utilities.utils import ensure_file_exists
from approvaltests.core.reporter import Reporter
from approvaltests.reporters.clipboard_reporter import get_command_text

#!/usr/bin/env python


class PythonNativeReporter(Reporter):
    """
    A reporter that outputs diff straight
    to standard output.

    This is useful when running in a non-GUI environment,
    such as in Continuous Integration systems.
    """

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        ensure_file_exists(approved_path)
        print(calculate_diff_with_approve_instruction(received_path, approved_path))
        return True

    @override
    def __str__(self) -> str:
        return self.__class__.__name__

    __repr__ = __str__


def calculate_diff_with_approve_instruction(file1: str, file2: str) -> str:
    diff_string = calculate_diff(file1, file2)
    if diff_string.strip():
        approve = get_command_text(file1, file2)
        approve_cmd = "\n\nto approve this result:\n\n" + approve + "\n"
    else:
        approve_cmd = ""
    return diff_string + approve_cmd


def calculate_diff(filename1: str, filename2: str) -> str:
    with open(filename1, encoding="utf8") as file1:
        with open(filename2, encoding="utf8") as file2:
            diff = unified_diff(
                file2.readlines(),
                file1.readlines(),
                os.path.basename(filename2),
                os.path.basename(filename1),
            )
            diff_string = "".join(diff)
            return diff_string


if __name__ == "__main__":
    print(calculate_diff_with_approve_instruction(sys.argv[1], sys.argv[2]))



import abc
import typing

from typing_extensions import override


class FormatWrapper(abc.ABC):
    @abc.abstractmethod
    def wrap(self, data: typing.Any) -> typing.Any:
        pass

    @abc.abstractmethod
    def is_match(self, data: typing.Any) -> bool:
        pass


class AlwaysMatch(FormatWrapper):
    @override
    def wrap(self, data: typing.Any) -> typing.Any:
        return data

    @override
    def is_match(self, data: typing.Any) -> bool:
        return True



import re
from collections import defaultdict
from typing import Callable, DefaultDict, Union

from approval_utilities.utilities.logger.logging_instance import print_type

Scrubber = Callable[[str], str]


# def print_type(value):
#     return f"<{type(value).__name__}>"
def create_regex_scrubber(
    regex: str, function_or_replace_string: Union[Callable[[int], str], str]
) -> Scrubber:
    if isinstance(function_or_replace_string, str):
        return lambda t: _replace_regex(t, regex, lambda _: function_or_replace_string)
    elif callable(function_or_replace_string):
        return lambda t: _replace_regex(t, regex, function_or_replace_string)
    else:
        raise TypeError(
            f"Parameter function_or_replace_string expects a string or callable, but got {print_type(function_or_replace_string)}"
        )


def _replace_regex(text: str, regex: str, replacement: Callable[[int], str]) -> str:
    matches: DefaultDict[str, int] = defaultdict(lambda: len(matches))
    return re.sub(regex, lambda m: replacement(matches[m.group(0)]), text)


def scrub_all_dates(date: str) -> str:
    return create_regex_scrubber(
        r"\d{4}-\d{2}-\d{2}\ \d{2}:\d{2}:\d{2}", lambda t: f"<date{t}>"
    )(date)


def scrub_all_guids(data: str) -> str:
    return create_regex_scrubber(
        r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}",
        lambda t: f"<guid_{t}>",
    )(data)


def combine_scrubbers(*scrubbers: Scrubber) -> Scrubber:
    def combined(data: str) -> str:
        for scrubber in scrubbers:
            data = scrubber(data)
        return data

    return combined


def templates_regex_scrubber_with_lambda() -> Scrubber:
    """
    This method exists as a convenient way to get an example scrubber for you to use.
    To use this template, simply inline the method in your IDE.
    """
    return create_regex_scrubber(
        "your pattern here: [a-zA-Z]+/d{4}", lambda t: f"<your replacement_{t}>"
    )


def templates_regex_scrubber_with_replacement() -> Scrubber:
    """
    This method exists as a convient way to get an example scrubber for you to use.
    To use this template, simply inline the method in your IDE.
    """
    return create_regex_scrubber(
        "your pattern here: [a-zA-Z]+/d{4}", "<your replacement>"
    )


def create_line_scrubber(remove_lines_containing: str) -> Scrubber:
    return lambda t: scrub_lines_containing(remove_lines_containing, t)


def scrub_lines_containing(remove_lines_containing: str, text: str) -> str:
    lines = text.splitlines()
    for line in lines:
        if remove_lines_containing in line:
            lines.remove(line)
    return "\n".join(lines)



from pathlib import Path


def test_init_present() -> None:
    current_working_directory = Path(Path(__file__).parents[1])
    assert_directories_have_inits(current_working_directory / "approvaltests")
    assert_directories_have_inits(current_working_directory / "approval_utilities")


def is_generated_code(directory_name: str) -> bool:
    return "__" in directory_name


def assert_directories_have_inits(directory: Path) -> None:
    if not is_generated_code(f"{directory}"):
        if not (directory / "__init__.py").exists():
            assert False, f"__init__.py is missing from {directory}"
        for subdirectory in directory.iterdir():
            if subdirectory.is_dir():
                assert_directories_have_inits(subdirectory)



import shutil

from typing_extensions import override

from approvaltests.core import Reporter


class ReporterThatAutomaticallyApproves(Reporter):
    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        shutil.move(received_path, approved_path)
        return True



from pathlib import Path

from approvaltests.internals.logs.log_commons import (
    APPROVAL_TESTS_TEMP_DIRECTORY,
    LogCommons,
)


class ApprovedFilesLog:
    @staticmethod
    def clear_log_file() -> None:
        ApprovedFilesLog.get_approved_files_log().write_text("")
        LogCommons.download_script_from_common_repo_if_needed(
            "remove_abandoned_files.py"
        )

    @staticmethod
    def get_approved_files_log() -> Path:
        return ApprovedFilesLog.get_temp_directory() / ".approved_files.log"

    @staticmethod
    def get_temp_directory() -> Path:
        APPROVAL_TESTS_TEMP_DIRECTORY.mkdir(parents=True, exist_ok=True)
        APPROVAL_TESTS_TEMP_DIRECTORY.joinpath(".gitignore").write_text("*")
        return APPROVAL_TESTS_TEMP_DIRECTORY

    @staticmethod
    def log(approved_file: str) -> None:
        with ApprovedFilesLog.get_approved_files_log().open(mode="a") as file:
            file.write(f"{approved_file}\n")


ApprovedFilesLog.clear_log_file()



from approvaltests.reporters.first_working_reporter import FirstWorkingReporter
from approvaltests.reporters.report_with_beyond_compare import ReportWithBeyondCompare


class ReportWithDiffToolOnWindows(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(ReportWithBeyondCompare())



import fnmatch
import inspect
import os
from inspect import FrameInfo
from typing import Dict, List, Optional

from typing_extensions import override

from approval_utilities.utilities.stack_frame_utilities import get_class_name_for_frame
from approvaltests.approval_exception import FrameNotFound
from approvaltests.integrations.pytest.pytest_config import PytestConfig
from approvaltests.namer.namer_base import NamerBase


class StackFrameNamer(NamerBase):
    directory = ""
    method_name = ""
    class_name = ""

    def __init__(self, extension: Optional[str] = None) -> None:
        NamerBase.__init__(self, extension)
        self.set_for_stack(inspect.stack(1))
        self.config: Dict[str, str] = {}
        self.config_loaded = False

    def set_for_stack(self, caller: List[FrameInfo]) -> None:
        frame = self.get_test_frame_index(caller)
        stacktrace = caller[frame]
        self.method_name = stacktrace[3]
        self.class_name = get_class_name_for_frame(stacktrace)
        self.directory = os.path.dirname(stacktrace[1])

    @staticmethod
    def get_test_frame_index(caller: List[FrameInfo]) -> int:
        tmp_array = []
        for index, frame in enumerate(caller):
            if StackFrameNamer.is_test_method(frame):
                tmp_array.append(index)
        if tmp_array:
            return tmp_array[-1]
        message = """Could not find test method/function. Possible reasons could be:
1) approvaltests is not being used inside a test function
2) your test framework is not supported by ApprovalTests (unittest and pytest are currently supported)."""
        raise FrameNotFound(message)

    @staticmethod
    def is_pytest_test(frame: FrameInfo) -> bool:
        method_name = frame[3]
        patterns = PytestConfig.test_naming_patterns
        return StackFrameNamer._is_match_for_pytest(method_name, patterns)

    @staticmethod
    def _is_match_for_pytest(method_name: str, patterns: List[str]) -> bool:
        # Do not modify this method, so we can compare with original code
        # taken from pytest/python.py (class PyCollector)
        for pattern in patterns:
            if method_name.startswith(pattern):
                return True
                # Check that name looks like a glob-string before calling fnmatch
                # because this is called for every name in each collected module,
                # and fnmatch is somewhat expensive to call.
            elif (
                "*" in pattern or "?" in pattern or "[" in pattern
            ) and fnmatch.fnmatch(method_name, pattern):
                return True

        return False

    @staticmethod
    def is_unittest_test(frame: FrameInfo) -> bool:
        method_name = frame[3]
        local_attributes = frame[0].f_locals
        is_unittest_test = (
            "self" in local_attributes
            and hasattr(local_attributes["self"], "__dict__")
            and "_testMethodName" in vars(local_attributes["self"])
            and method_name != "__call__"
            and method_name != "_callTestMethod"
            and method_name != "run"
        )
        return is_unittest_test

    @staticmethod
    def is_test_method(frame: FrameInfo) -> bool:
        return StackFrameNamer.is_unittest_test(
            frame
        ) or StackFrameNamer.is_pytest_test(frame)

    def get_class_name(self) -> str:
        return self.class_name

    def get_method_name(self) -> str:
        return self.method_name

    @override
    def get_directory(self) -> str:
        return self.directory

    @override
    def config_directory(self) -> str:
        return self.directory

    @override
    def get_file_name(self) -> str:
        class_name = "" if (self.class_name is None) else (self.class_name + ".")
        return class_name + self.method_name

    def get_extension_with_dot(self) -> str:
        return self.extension_with_dot

    def get_extension_without_dot(self) -> str:
        return self.extension_with_dot[1:]

    @classmethod
    def get_test_frame(cls) -> FrameInfo:
        calling_stack = inspect.stack(1)
        frame = StackFrameNamer.get_test_frame_index(calling_stack)
        return calling_stack[frame]



import unittest
from pathlib import Path
from typing import Optional

from typing_extensions import override

from approval_utilities.utils import get_adjacent_file
from approvaltests import Options, assert_against_file, assert_equal_with_reporter
from approvaltests.approval_exception import ApprovalException
from approvaltests.core import Reporter
from approvaltests.reporters.testing_reporter import ReporterForTesting
from approvaltests.scrubbers import scrub_all_guids


class TestAssertEqualWithReporter(unittest.TestCase):
    def test_assert_with_scrubbing_and_options(self) -> None:
        actuals = "2fd78d4a-ad49-447d-96a8-deda585a9aa5 and text"
        expected = "<guid_0> and text"
        assert_equal_with_reporter(
            expected,
            actuals,
            options=Options()
            .with_scrubber(scrub_all_guids)
            .for_file.with_extension(".md"),
        )

    def test_text_reporter_called_on_failure(self) -> None:
        class LocalReporter(Reporter):
            def __init__(self) -> None:
                self.received: Optional[str] = None
                self.approved: Optional[str] = None
                self.extention: Optional[str] = None

            @override
            def report(self, received_path: str, approved_path: str) -> bool:
                self.received = Path(received_path).read_text(encoding="utf-8-sig")
                self.approved = Path(approved_path).read_text(encoding="utf-8-sig")
                self.extention = Path(received_path).suffix

        reporter = LocalReporter()
        try:
            assert_equal_with_reporter(
                "expected",
                "actual",
                options=Options()
                .with_reporter(reporter)
                .for_file.with_extension(".md"),
            )
        except AssertionError:
            pass
        self.assertEqual(reporter.received, "actual")
        self.assertEqual(reporter.approved, "expected")
        self.assertEqual(reporter.extention, ".md")


class TestAsserts(unittest.TestCase):
    def test_assert_against_file_works(self) -> None:
        file_path = get_adjacent_file("manual_file.approved.txt")
        assert_against_file("This text is in a file\n", file_path)

    def test_assert_against_file_fails_with_reporter(self) -> None:
        reporter = ReporterForTesting()
        file_path = get_adjacent_file("manual_file.approved.txt")
        try:
            assert_against_file("This text is NOT in a file", file_path, reporter)
        except ApprovalException:
            pass
        self.assertTrue(reporter.called)



from typing_extensions import override

from approvaltests.core import Reporter


class ReportQuietly(Reporter):
    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        return True


QuietReport = ReportQuietly  # backwards compatibility



from typing_extensions import override

from approvaltests.core.reporter import Reporter


class ReporterForTesting(Reporter):
    def __init__(self) -> None:
        self.called = False

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        self.called = True
        return True



from .generic_diff_reporter import GenericDiffReporter, create_config


class ReportWithDiffCommandLine(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=create_config(["ReportWithDiffCommandLine", "diff", ["-u"]])
        )



import io
from pathlib import Path

import numpy as np
from typing_extensions import override

from approvaltests import Options, verify_binary
from approvaltests.core import Reporter
from approvaltests.reporters import get_command_text


def serialize_ndarray(a: np.ndarray) -> bytes:
    bs = io.BytesIO()
    np.save(bs, a)
    bs.seek(0)
    return bs.read()


# begin-snippet: verify_numpy_array
def test_simulator_produces_correct_output() -> None:
    np_array = np.full(shape=(32, 16), fill_value=42, dtype=np.int64)
    verify_binary(
        serialize_ndarray(np_array),
        ".npy",
        options=Options().with_reporter(NDArrayDiffReporter()),
    )


# end-snippet

# Use a custom reporter to show the assertion message from numpy
# Also the default python console reporter will try to interpret the numpy data as text causing an error

# begin-snippet: numpy_custom_reporter


def load_ndarray(path: str):
    with open(path, mode="rb") as f:
        return np.load(f)


class NDArrayDiffReporter(Reporter):
    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        if not Path(approved_path).is_file():
            self._create_empty_array(approved_path)
        received = load_ndarray(received_path)
        approved = load_ndarray(approved_path)
        to_approve_msg = (
            f"To approve run:\n {get_command_text(received_path, approved_path)}"
        )
        print(np.testing.build_err_msg([received, approved], err_msg=to_approve_msg))
        return True

    # end-snippet

    @staticmethod
    def _create_empty_array(path: str) -> None:
        with open(path, mode="wb") as f:
            f.write(serialize_ndarray(np.zeros((0,))))



from approvaltests.reporters.first_working_reporter import FirstWorkingReporter
from approvaltests.reporters.generic_diff_reporter import GenericDiffReporter
from approvaltests.reporters.generic_diff_reporter_config import (
    GenericDiffReporterConfig,
)


class ReportWithBeyondCompareLinux(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__, path="/usr/bin/bcompare"
            )
        )


class ReportWithBeyondCompareMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/Beyond Compare.app/Contents/MacOS/BCompare",
            )
        )


class ReportWithBeyondCompare4Windows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}/Beyond Compare 4/BCompare.exe",
            )
        )


class ReportWithBeyondCompare5Windows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}/Beyond Compare 5/BCompare.exe",
            )
        )


class ReportWithWinMerge(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}/WinMerge/WinMergeU.exe",
            )
        )


class ReportWithPycharm(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}/JetBrains/PyCharm 2021.2.2/bin/pycharm64.exe",
                extra_args=["diff"],
            )
        )


class ReportWithBeyondCompare(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithBeyondCompareMac(),
            ReportWithBeyondCompare4Windows(),
            ReportWithBeyondCompare5Windows(),
            ReportWithBeyondCompareLinux(),
        )


def report_with_beyond_compare() -> ReportWithBeyondCompare:
    return ReportWithBeyondCompare()



from typing import Optional

from typing_extensions import override

from approval_utilities.utilities.clipboard_utilities import copy_to_clipboard
from approval_utilities.utils import is_windows_os
from approvaltests.core.reporter import Reporter


def get_command_text(
    received_path: str, approved_path: str, is_windows: Optional[bool] = None
) -> str:
    if is_windows is None:
        is_windows = is_windows_os()
    if is_windows:
        command = f'move "{received_path}" "{approved_path}"'
    else:
        command = f"mv -f {received_path} {approved_path}"
    return command


class ClipboardReporter(Reporter):
    """
    A reporter that creates
    a command line suitable for approving
    the last failed test on systems that
    support terminal command 'mv', and puts
    the command on the clipboard, over-writes
    the previous clipboard contents.

    See also CommandLineReporter.
    """

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        text = get_command_text(received_path, approved_path)
        print(text)
        copy_to_clipboard(text)
        return True


class CommandLineReporter(Reporter):
    """
    A reporter that outputs a
    command line suitable for approving
    failing tests on systems that support
    terminal command 'mv'.

    The output is typically copied and pasted
    to a console window or script, to approve
    the new test results.

    See also ClipboardReporter.
    """

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        text = get_command_text(received_path, approved_path)
        print(f"\n\n{text}\n\n")
        return True



import json
import random
import unittest
from typing import Callable, Dict, Optional

import pytest
from typing_extensions import override

from approval_utilities.approvaltests.core.verifiable import Verifiable
from approval_utilities.approvaltests.core.verify_parameters import VerifyParameters
from approval_utilities.utilities.multiline_string_utils import remove_indentation_from
from approval_utilities.utils import get_adjacent_file, is_windows_os, print_grid
from approvaltests import List, Options, approvals, delete_approved_file
from approvaltests.approval_exception import ApprovalException
from approvaltests.approvals import (
    verify,
    verify_as_json,
    verify_binary,
    verify_exception,
    verify_file,
    verify_html,
    verify_xml,
)
from approvaltests.core.comparator import Comparator
from approvaltests.reporters.report_all_to_clipboard import (
    ReporterByCopyMoveCommandForEverythingToClipboard,
)
from approvaltests.reporters.report_with_beyond_compare import ReportWithPycharm
from approvaltests.reporters.reporter_that_automatically_approves import (
    ReporterThatAutomaticallyApproves,
)
from approvaltests.reporters.testing_reporter import ReporterForTesting
from approvaltests.storyboard import Storyboard, verify_storyboard


class GameOfLife:
    def __init__(self, board: Callable[[int, int], int]):
        self.board = board
        self.alive = "X"
        self.dead = "."

    def advance(self) -> "GameOfLife":
        old = self.board

        def my_next(x, y):
            count = (
                old(x + 1, y)
                + old(x + 1, y - 1)
                + old(x + 1, y + 1)
                + old(x - 1, y)
                + old(x - 1, y - 1)
                + old(x - 1, y + 1)
                + old(x, y + 1)
                + old(x, y - 1)
            )
            return count == 3 or (count == 2 and old(x, y))

        self.board = my_next
        return self

    @override
    def __str__(self):
        return print_grid(
            5, 5, lambda x, y: f"{self.alive} " if self.board(x, y) else f"{self.dead} "
        )

    def set_alive_cell(self, alive: str) -> str:
        self.alive = alive
        return self.alive

    def set_dead_cell(self, dead: str) -> str:
        self.dead = dead
        return self.dead


class VerifyTests(unittest.TestCase):
    @override
    def setUp(self) -> None:
        self.reporter = ReporterByCopyMoveCommandForEverythingToClipboard()

    def test_verify(self) -> None:
        verify("Hello World.", self.reporter)

    def test_verify_with_encoding(self) -> None:
        verify(
            "Høvdingens kjære squaw får litt pizza i Mexico by",
            self.reporter,
            encoding="utf-8",
        )

    def test_verify_with_encoding_error_raises_value_error(self) -> None:
        with self.assertRaises(ValueError):
            verify(
                "Høvdingens kjære squaw får litt pizza i Mexico by",
                self.reporter,
                encoding="ascii",
            )

    def test_verify_with_errors_replacement_character(self) -> None:
        verify(
            "Falsches Üben von Xylophonmusik quält jeden größeren Zwerg",
            self.reporter,
            encoding="ascii",
            errors="replace",
        )

    def test_verify_with_newlines(self) -> None:
        verify(
            "I cannot live without approval.\n"
            "Your satisfaction is my demand.\n"
            "I must control what you think of me.\n"
            "I have to understand.\n",
            reporter=self.reporter,
            encoding="utf-8",
            newline="\r\n",
        )

    def test_verify_fail(self) -> None:
        reporter = ReporterForTesting()
        try:
            verify("Hello World.", reporter)
            self.assertFalse(True, "expected exception")
        except ApprovalException as e:
            self.assertTrue("Approval Mismatch", e.value)

    def test_verify_as_json(self) -> None:
        class Bag:
            def __init__(self) -> None:
                self.stuff = 1
                self.json: Optional[Dict[str, int]] = None

        o = Bag()
        o.json = {"a": 0, "z": 26}
        verify_as_json(o)

    def test_verify_as_json_raises_type_error_for_non_renderable_types(self):
        with self.assertRaises(AttributeError):
            verify_as_json(Ellipsis, self.reporter)

    def test_verify_as_json_raises_value_error_for_non_renderable_values(self):
        circular_data: List[List] = []
        circular_data.append(circular_data)
        with self.assertRaises(ValueError):
            verify_as_json(circular_data, self.reporter)

    def test_json_in_json(self):
        dict = {"a": 1, "b": 2, "c": 3}
        verify_as_json(
            {"type": "dictionary", "value": json.dumps(dict)},
            deserialize_json_fields=True,
        )

    def test_json_in_dict_in_json(self):
        dict = {"a": 1, "b": 2, "c": 3}
        verify_as_json(
            {"type": "dictionary", "value": {"key": json.dumps(dict)}},
            deserialize_json_fields=True,
        )

    def test_verify_file(self) -> None:
        name = "exampleFile.txt"
        filename = get_adjacent_file(name)
        verify_file(filename, self.reporter)

    @pytest.mark.skipif(not is_windows_os(), reason="Doesn't work on unix")
    def test_verify_file_with_windows_1252_encoding(self):
        name = "exampleFileWindows1252.txt"
        filename = get_adjacent_file(name)
        verify_file(filename, self.reporter)

    def test_verify_file_with_actual_windows_1252_encoding(self) -> None:
        name = "exampleFile_Actual_Windows1252.txt"
        filename = get_adjacent_file(name)
        verify_file(filename, self.reporter)

    def test_verify_file_binary_file(self) -> None:
        name = "icon.png"
        filename = get_adjacent_file(name)
        verify_file(filename)

    def test_verify_bytes(self) -> None:
        # begin-snippet: verify_binary_image
        name = "icon.png"
        filename = get_adjacent_file(name)
        with open(filename, mode="rb") as f:
            verify_binary(f.read(), ".png")
        # end-snippet

    def test_verify_xml(self) -> None:
        xml = """<?xml version="1.0" encoding="UTF-8"?><orderHistory createdAt='2019-08-02T16:40:18.109470'><order date='2018-09-01T00:00:00+00:00' totalDollars='149.99'><product id='EVENT02'>Makeover</product></order><order date='2017-09-01T00:00:00+00:00' totalDollars='14.99'><product id='LIPSTICK01'>Cherry Bloom</product></order></orderHistory>"""
        verify_xml(xml)

    def test_verify_html(self) -> None:
        html = """<!DOCTYPE html><html><head> <title>Example</title> </head> <body> <p>This is an example of a simple HTML page with one paragraph.</p></body></html>"""
        verify_html(html)

    def test_newlines_at_end_of_files(self) -> None:
        verify(
            "There should be a blank line underneath this",
            options=Options().with_reporter(ReportWithPycharm()),
        )

    def test_verify_storyboard(self) -> None:
        with verify_storyboard() as b:
            game_of_life = GameOfLife(lambda x, y: 2 <= x <= 4 and y == 2)
            b.add_frame(game_of_life)
            b.add_frames(2, lambda _: game_of_life.advance())

    def test_storyboard(self) -> None:
        game_of_life = GameOfLife(lambda x, y: 2 <= x <= 4 and y == 2)
        verify(
            Storyboard()
            .add_frame(game_of_life)
            .add_frames(2, lambda _: game_of_life.advance())
        )

    def test_simple_storyboard(self) -> None:
        class AsciiWheel:
            def __init__(self) -> None:
                self.steps = ["-", "\\", "|", "/"]
                self.step = 0

            @override
            def __str__(self) -> str:
                return self.steps[self.step]

            def advance(self) -> None:
                self.step += 1
                self.step %= 4

        ascii_wheel = AsciiWheel()
        # begin-snippet: use_storyboard
        story = Storyboard()
        story.add_description("Spinning wheel")
        story.add_frame(ascii_wheel)
        ascii_wheel.advance()
        story.add_frame(ascii_wheel)
        verify(story)
        # end-snippet

    def test_storyboard_of_iterable(self) -> None:
        approvals.settings().allow_multiple_verify_calls_for_this_method()
        spinning_wheel = ["-", "\\", "|", "/"] * 3
        verify(Storyboard().iterate_frames(spinning_wheel, 5))

        spinning_wheel = ["-", "\\", "|", "/", "-"]
        verify(Storyboard().iterate_frames(spinning_wheel))

    def test_other_storyboard_machanisms(self) -> None:
        game_of_life = GameOfLife(lambda x, y: 1 <= x <= 3 and y == 2)

        story = Storyboard()
        story.add_description("Game of Life")
        story.add_frame(game_of_life)

        game_of_life = game_of_life.advance()
        story.add_frame(game_of_life, "Start game_of_life")

        game_of_life = game_of_life.advance()
        story.add_frame(game_of_life)

        story.add_description_with_data(
            "setting alive", game_of_life.set_alive_cell("*")
        )
        story.add_description_with_data("setting dead", game_of_life.set_dead_cell("_"))
        game_of_life = game_of_life.advance()
        story.add_frame(game_of_life)

        game_of_life = game_of_life.advance()
        story.add_frame(game_of_life)

        game_of_life = game_of_life.advance()
        story.add_frame(game_of_life)

        verify(story)

    def test_exist_file_extension(self):
        verify_file(get_adjacent_file("sample.xml"))

    def test_exist_file_with_modified_extension(self):
        verify_file(
            get_adjacent_file("sample.xml"),
            options=Options().for_file.with_extension(".json"),
        )

    def test_verify_converts_to_string(self):
        verify(1)

    def test_verify_automatic_approval(self):
        approvals.settings().allow_multiple_verify_calls_for_this_method()
        delete_approved_file()
        with pytest.raises(ApprovalException):
            verify(
                2,
                options=Options().with_reporter(
                    reporter=ReporterThatAutomaticallyApproves()
                ),
            )
        verify(2)

    def test_verify_custom_comparator_allows_all_inputs(self):
        class EverythingIsTrue(Comparator):
            @override
            def compare(self, received_path: str, approved_path: str) -> bool:
                return True

        verify(random.random(), options=Options().with_comparator(EverythingIsTrue()))

    # begin-snippet: verifiable_object_example
    def test_verifiable(self):
        class MarkdownParagraph(Verifiable):
            def __init__(self, title: str, text: str) -> None:
                self.title = title
                self.text = text

            @override
            def __str__(self) -> str:
                return remove_indentation_from(
                    f""" 
                # {self.title}
                {self.text}
                """
                )

            @override
            def get_verify_parameters(self, options: Options) -> VerifyParameters:
                return VerifyParameters(options.for_file.with_extension(".md"))

        verify(
            MarkdownParagraph("Paragraph Title", "This is where the paragraph text is.")
        )

    # end-snippet

    def test_verify_exception(self):
        def throw_exception():
            raise RuntimeError("some error")

        verify_exception(throw_exception)



from pathlib import Path

import requests

APPROVAL_TESTS_TEMP_DIRECTORY = Path(".approval_tests_temp")


class LogCommons:
    @staticmethod
    def download_script_from_common_repo_if_needed(
        script_name_with_suffix: str,
    ) -> None:
        try:
            script_path = APPROVAL_TESTS_TEMP_DIRECTORY / script_name_with_suffix
            if script_path.exists():
                return

            response = requests.get(
                f"https://raw.githubusercontent.com/approvals/ApprovalTests.CommonScripts/refs/heads/main/{script_name_with_suffix}"
            )
            if response.ok:
                script_path.write_text(response.text)

                make_executable = 0o755
                script_path.chmod(make_executable)
        except:
            pass



from types import TracebackType
from typing import ContextManager, Optional, Type

from approval_utilities.utilities.logger.simple_logger import SimpleLogger
from approvaltests.approvals import verify
from approvaltests.core.options import Options


def verify_simple_logger(
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> ContextManager[None]:
    class VerifySimpleLogger:
        def __init__(self) -> None:
            self.output = SimpleLogger.log_to_string()

        def __enter__(self) -> None:
            pass

        def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType],
        ) -> bool:
            verify(self.output, options=options)

    return VerifySimpleLogger()



import json
from typing import Callable, Dict, Iterator, List, Optional

from typing_extensions import override

from approval_utilities.utilities.deprecated import deprecated
from approval_utilities.utils import get_adjacent_file
from approvaltests.core.reporter import Reporter
from approvaltests.reporters.generic_diff_reporter import GenericDiffReporter
from approvaltests.reporters.generic_diff_reporter_config import (
    GenericDiffReporterConfig,
    create_config,
)
from approvaltests.reporters.python_native_reporter import PythonNativeReporter
from approvaltests.reporters.report_with_beyond_compare import (
    ReportWithBeyondCompare,
    ReportWithWinMerge,
)


class NoConfigReporter(Reporter):
    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        raise RuntimeError("This machine has no reporter configuration")


class GenericDiffReporterFactory:
    reporter_configs: List[GenericDiffReporterConfig] = []

    def __init__(self) -> None:
        self.load(get_adjacent_file("reporters.json"))

    def add_default_reporter_config(self, config: List[str]) -> None:
        self.reporter_configs.insert(0, create_config(config))

    def list(self) -> List[str]:
        return [r.name for r in self.reporter_configs]

    def get(self, reporter_name: str) -> Reporter:
        reporter = GenericDiffReporterFactory.get_reporter_programmatically(
            reporter_name
        )
        return reporter or self.get_from_json_config(reporter_name)

    @staticmethod
    def get_reporter_programmatically(reporter_name: str) -> Optional[Reporter]:
        reporters: Dict[str, Callable[[], Reporter]] = {
            "BeyondCompare": ReportWithBeyondCompare,
            "WinMerge": ReportWithWinMerge,
            "PythonNative": PythonNativeReporter,
            "PythonNativeReporter": PythonNativeReporter,
        }
        clazz = reporters.get(reporter_name)
        return clazz and clazz()

    def get_from_json_config(self, reporter_name: str) -> Reporter:
        config = next(
            (r for r in self.reporter_configs if r.name == reporter_name), None
        )
        if not config:
            return NoConfigReporter()
        return self._create_reporter(config)

    @staticmethod
    def _create_reporter(config: GenericDiffReporterConfig) -> GenericDiffReporter:
        return GenericDiffReporter(config)

    def save(self, file_name: str) -> str:
        with open(file_name, "w", encoding="utf8") as file:
            json.dump(
                [reporter.serialize() for reporter in self.reporter_configs],
                file,
                sort_keys=True,
                indent=2,
                separators=(",", ": "),
            )
        return file_name

    def load(self, file_name: str) -> List[GenericDiffReporterConfig]:
        with open(file_name, encoding="utf8") as file:
            configs = json.load(file)
        self.reporter_configs = [create_config(config) for config in configs]
        return self.reporter_configs

    def get_first_working(self) -> Optional[GenericDiffReporter]:
        working = (i for i in self.get_all_reporters_from_config() if i.is_working())
        return next(working, None)

    @deprecated(reason="Please use get_all_reporters_from_json()")
    def get_all_reporters(self) -> Iterator[GenericDiffReporter]:
        return self.get_all_reporters_from_config()

    def get_all_reporters_from_config(self) -> Iterator[GenericDiffReporter]:
        instances = (self._create_reporter(r) for r in self.reporter_configs)
        return instances

    def remove(self, reporter_name: str) -> None:
        self.reporter_configs = [
            r for r in self.reporter_configs if r.name != reporter_name
        ]



import platform
import subprocess
from pathlib import Path
from typing import List

from typing_extensions import override

from approval_utilities.utilities.logger.simple_logger import SimpleLogger
from approvaltests.core.reporter import Reporter


class ReportByOpeningFiles(Reporter):
    @staticmethod
    def display_file(file_path: str) -> None:
        call = ReportByOpeningFiles.get_opening_command(file_path, platform.system())
        subprocess.call(call, shell=True)

    @staticmethod
    def get_opening_command(file_path: str, os: str) -> List[str]:
        command = {
            "Windows": "start",
            "Darwin": "open",
            "Linux": "xdg-open",
        }[os]
        call = [command, file_path]
        return call

    @staticmethod
    def is_non_empty_file(file_path: str) -> bool:
        path = Path(file_path)
        return path.is_file() and path.stat().st_size > 0

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        try:
            if self.is_non_empty_file(approved_path):
                self.display_file(approved_path)

            self.display_file(received_path)

            return True
        except Exception as e:
            SimpleLogger.warning("Failed to open files", exception=e)
            return False



from typing_extensions import override

from approval_utilities.utilities.clipboard_utilities import copy_to_clipboard
from approvaltests.core.reporter import Reporter
from approvaltests.reporters.clipboard_reporter import get_command_text


class ReporterByCopyMoveCommandForEverythingToClipboard(Reporter):
    text: str = ""

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        ReporterByCopyMoveCommandForEverythingToClipboard.text = (
            ReporterByCopyMoveCommandForEverythingToClipboard.text
            + get_command_text(received_path, approved_path)
            + "\n"
        )
        copy_to_clipboard(ReporterByCopyMoveCommandForEverythingToClipboard.text)
        return True



import filecmp
import os
import pathlib
from typing import Callable, List, Optional

from typing_extensions import override

from approval_utilities.utilities.multiline_string_utils import remove_indentation_from
from approvaltests.core.comparator import Comparator
from approvaltests.core.namer import Namer
from approvaltests.core.reporter import Reporter
from approvaltests.core.writer import Writer
from approvaltests.internals.logs.approved_file_log import ApprovedFilesLog
from approvaltests.internals.logs.failed_comparison_log import FailedComparisonLog


def exists(path: str) -> bool:
    return os.path.isfile(path)


class ReporterNotWorkingException(Exception):
    def __init__(self, reporter: Reporter):
        super().__init__(f"Reporter {reporter} failed to work!")


class FileComparator(Comparator):
    @override
    def compare(self, received_path: str, approved_path: str) -> bool:
        if not exists(approved_path) or not exists(received_path):
            return False
        if filecmp.cmp(approved_path, received_path, shallow=False):
            return True
        try:
            approved_raw = pathlib.Path(approved_path).read_text()
            approved_text = approved_raw.replace("\r\n", "\n")
            received_raw = pathlib.Path(received_path).read_text()
            received_text = received_raw.replace("\r\n", "\n")
            return approved_text == received_text
        except BaseException:
            return False


class FileApprover:
    previous_approved: List[str] = []
    allowed_duplicates: List[Callable[[str], bool]] = []

    @staticmethod
    def verify(
        namer: Namer, writer: Writer, reporter: Reporter, comparator: Comparator
    ) -> Optional[str]:
        approved = namer.get_approved_filename()
        received = namer.get_received_filename()

        if FileApprover.is_this_a_multiple_verify(approved):
            return FileApprover.get_duplicate_verify_error_message(approved)
        FileApprover.previous_approved.append(approved)

        # The writer has the ability to change the name of the received file
        received = writer.write_received_file(received)
        verified = FileApprover.verify_files(approved, received, reporter, comparator)

        if not verified:
            return (
                f"Approval Mismatch, received != approved\n"
                f"\tApproved: {approved}\n"
                f"\tReceived: {received} "
            )
        return None

    @staticmethod
    def get_duplicate_verify_error_message(approved: str) -> str:
        return remove_indentation_from(
            f"""
            We noticed that you called verify more than once in the same test. 
            This is the second call to verify:
                approved_file: {approved}
    
            By default, ApprovalTests only allows one verify() call per test.
            To find out more, visit: 
            https://github.com/approvals/ApprovalTests.Python/blob/main/docs/how_to/multiple_approvals_per_test.md
            
            # Possible Fixes
            1. Separate your test into two tests
            2. In your verify call, add `options=NamerFactory.with_parameters("your_paramater")`
            3. In your test, call `approvals.settings().allow_multiple_verify_calls_for_this_method()`
            """
        )

    @staticmethod
    def is_this_a_multiple_verify(approved: str) -> bool:
        return (
            approved in FileApprover.previous_approved
            and not FileApprover.is_duplicate_allowed(approved)
        )

    @staticmethod
    def is_duplicate_allowed(approved: str) -> bool:
        for allowed in FileApprover.allowed_duplicates:
            if allowed(approved):
                return True
        return False

    @staticmethod
    def verify_files(
        approved_file: str,
        received_file: str,
        reporter: Reporter,
        comparator: Comparator,
    ) -> bool:
        ApprovedFilesLog.log(approved_file)

        if comparator.compare(received_file, approved_file):
            os.remove(received_file)
            return True
        FailedComparisonLog.log(received_file, approved_file)
        worked = reporter.report(received_file, approved_file)
        if not worked:
            raise ReporterNotWorkingException(reporter)
        return False

    @staticmethod
    def add_allowed_duplicates(is_duplicate_allowed: Callable[[str], bool]) -> None:
        FileApprover.allowed_duplicates.append(is_duplicate_allowed)
